from io import BytesIO
import time

import nengo
import numpy as np


class CameraPipeline(object):
    def __init__(self, device='/dev/video0', width=640, height=480):
        import gi
        gi.require_version('Gst', '1.0')
        from gi.repository import Gst
        self.Gst = Gst

        self.device = device
        self.width = int(width)
        self.height = int(height)
        self.channels = 3
        self.size = self.width * self.height * self.channels
        assert self.size > 0

        self.running = False
        self.pipeline = None
        self.source = None
        self.parse = None
        self.sink = None

        self.time = -1
        self.jpeg = None

        self.init_gst()

    def init_gst(self):
        self.Gst.init(None)

        self.pipeline = self.Gst.Pipeline.new("camera")

        self.source = self.Gst.ElementFactory.make("v4l2src", "source")
        self.source.set_property("device", self.device)

        self.parse = self.Gst.ElementFactory.make("jpegparse", "parse")

        self.sink = self.Gst.ElementFactory.make("appsink", "sink")
        self.sink.set_property("drop", True)
        self.sink.set_property("max-buffers", 3)
        self.sink.set_property("emit-signals", True)
        self.sink.set_property("sync", False)
        self.sink.connect("new-sample", self.frame_from_sink)

        self.pipeline.add(self.source)
        self.pipeline.add(self.parse)
        self.pipeline.add(self.sink)

        caps = self.Gst.caps_from_string(
            'image/jpeg,width=%d,height=%d,framerate=30/1' %
            (self.width, self.height))
        self.source.link_filtered(self.parse, caps)
        self.parse.link(self.sink)

    def frame_from_sink(self, sink):
        sample = sink.emit("pull-sample")
        if sample is not None:
            buffer = sample.get_buffer()
            string = buffer.extract_dup(0, buffer.get_size())
            self.time = time.time()
            self.jpeg = string

        return False

    def start(self):
        if not self.running:
            self.running = True
            self.pipeline.set_state(self.Gst.State.PAUSED)
            self.pipeline.set_state(self.Gst.State.PLAYING)

    def stop(self):
        self.pipeline.set_state(self.Gst.State.NULL)


class CameraData(object):
    def __init__(self, width, height, channels=3, keep_aspect=True):
        import PIL.Image
        import PIL.ImageFile
        self.PILImage = PIL.Image

        PIL.ImageFile.LOAD_TRUNCATED_IMAGES = True

        self.width = width
        self.height = height
        self.channels = channels
        self.keep_aspect = keep_aspect

        self.time = -1
        self.jpeg = None
        self.array = None

    def set_jpeg(self, time, jpeg):
        self.time = time
        self.jpeg = jpeg

        buffer = BytesIO()
        buffer.write(jpeg)
        buffer.seek(0)
        image = self.PILImage.open(buffer)
        camshape = np.array([image.height, image.width])

        if self.keep_aspect:
            aspect = float(self.width) / self.height

            crop0 = min(camshape[0], camshape[1]/aspect)
            cropshape = np.array([int(crop0), int(crop0*aspect)])
            crop_a = (camshape - cropshape) // 2
            crop_b = crop_a + cropshape
            image = image.crop((crop_a[1], crop_a[0], crop_b[1], crop_b[0]))

        image = image.resize((self.width, self.height))
        array = np.asarray(image)
        array = np.transpose(array, (2, 0, 1))  # put color channel first
        self.array = array


class Camera(nengo.Process):
    def __init__(self, width=256, height=256, offset=-128,
                 device='/dev/video0', cam_width=640, cam_height=480):
        self.pipeline = CameraPipeline(
            device=device, width=cam_width, height=cam_height)

        self.width = int(width)
        self.height = int(height)
        self.channels = self.pipeline.channels
        self.size = self.width * self.height * self.channels
        assert self.size > 0

        self.offset = offset
        self.framerate = 0.1

        super(Camera, self).__init__(
            default_size_in=0, default_size_out=self.size)

    def make_step(self, shape_in, shape_out, dt, rng):
        assert shape_in == (0,)
        assert shape_out == (self.size,)

        # pipeline = CameraPipeline(self.device)
        pipeline = self.pipeline

        data = CameraData(self.width, self.height, self.channels)
        zero_image = np.zeros((self.channels, self.width, self.height))
        offset = self.offset
        framerate = self.framerate

        def step_camera(t):
            pipeline.start()

            if pipeline.time - data.time > framerate:
                data.set_jpeg(pipeline.time, pipeline.jpeg)

            array = data.array if data.array is not None else zero_image
            array = array + offset
            return array.ravel()

        return step_camera



import numpy as np
import pytest

import nengo_extras.matplotlib as ne_plt


def wave_image(shape, angle, color=None):
    assert len(shape) == 2

    i = np.linspace(-1, 1, shape[0])
    j = np.linspace(-1, 1, shape[1])
    I, J = np.meshgrid(i, j, indexing='ij')

    c, s = np.cos(angle), np.sin(angle)
    # X = c*I + s*J
    X = -c*J + s*I

    image = 0.5*np.cos(2*np.pi*X) + 0.5

    if color is not None:
        color = np.asarray(color)
        assert color.size == 3
        image = image[:, :, None] * color[None, None, :]

    return image


def test_imshow_gray(plt):
    image = wave_image((31, 32), 0.2)
    ne_plt.imshow(image)


def test_imshow_color(plt):
    image = wave_image((31, 32), 0.6, color=(1, 1, 0))
    ne_plt.imshow(image)


def test_imshow_color_vlim(plt):
    image = wave_image((31, 32), 0.6, color=(1, 1, 0))
    image = 2*image - 1
    ne_plt.imshow(image, vmin=-1, vmax=1)


@pytest.mark.parametrize('color', (False, True))
def test_tile(plt, rng, color):
    n = 200
    angles = rng.uniform(0, np.pi, size=n)
    colors = rng.uniform(size=(n, 3))

    genimg = lambda k: wave_image(
        (30, 31), angles[k], color=colors[k] if color else None)
    images = np.array([genimg(k) for k in range(n)])

    ne_plt.tile(images, grid=True, gridcolor=(1, 1, 0))


@pytest.mark.parametrize('color', (False, True))
def test_compare(plt, rng, color):
    d = 3
    n = 10
    angles = rng.uniform(0, np.pi, size=n)
    colors = rng.uniform(size=(n, 3))

    genimg = lambda k: wave_image(
        (30, 31),
        angles[k] + rng.uniform(-0.3, 0.3),
        color=colors[k] if color else None)
    image_sets = [np.array([genimg(k) for k in range(n)]) for _ in range(d)]

    ne_plt.compare(image_sets)



from nengo.utils.compat import pickle, PY2


if PY2:
    from cStringIO import StringIO
    from urllib import urlretrieve
    import Tkinter as tkinter
else:
    from io import StringIO
    from urllib.request import urlretrieve
    import tkinter


def cmp(a, b):  # same as python2's builtin cmp, not available in python3
    return (a > b) - (a < b)


def pickle_load(file, *args, **kwargs):
    if not PY2:
        kwargs.setdefault('encoding', 'latin1')
    return pickle.load(file, *args, **kwargs)


def pickle_load_bytes(file, *args, **kwargs):
    if not PY2:
        kwargs.setdefault('encoding', 'bytes')
    return pickle.load(file, *args, **kwargs)



import numpy as np
import pytest

from nengo_extras.data import (
    load_cifar10, load_cifar100, load_ilsvrc2012, load_mnist, load_svhn,
    one_hot_from_labels, spasafe_name, spasafe_names)
from nengo_extras.matplotlib import tile


@pytest.mark.slow
def test_load_cifar10(plt):
    (trainX, _), (_, _) = load_cifar10(n_train=1, n_test=0)
    trainX = trainX.reshape(-1, 3, 32, 32)
    trainX = np.transpose(trainX, (0, 2, 3, 1))
    tile(trainX)


@pytest.mark.slow
def test_load_cifar100(plt):
    (trainX, _), (_, _) = load_cifar100()
    trainX = trainX.reshape(-1, 3, 32, 32)
    trainX = np.transpose(trainX, (0, 2, 3, 1))
    tile(trainX)


@pytest.mark.slow
def test_load_ilsvrc2012(plt):
    testX, _, _, _ = load_ilsvrc2012(n_files=1)
    testX = testX.reshape(-1, 3, 256, 256)
    testX = np.transpose(testX, (0, 2, 3, 1))
    tile(testX)


@pytest.mark.slow
def test_load_mnist(plt):
    (trainX, _), (_, _) = load_mnist()
    tile(trainX.reshape(-1, 28, 28))


@pytest.mark.slow
def test_load_svhn(plt):
    (trainX, _), (_, _) = load_svhn()
    trainX = trainX.reshape(-1, 3, 32, 32)
    trainX = np.transpose(trainX, (0, 2, 3, 1))
    tile(trainX)


def test_one_hot_from_labels_int(rng):
    nc = 19
    labels = rng.randint(nc, size=1000)

    yref = np.zeros((len(labels), nc))
    yref[np.arange(len(labels)), labels] = 1

    y0 = one_hot_from_labels(labels)
    y1 = one_hot_from_labels(labels, classes=nc+5)
    assert np.array_equal(y0, yref)
    assert np.array_equal(y0, y1[:, :nc])
    assert (y1[:, nc:] == 0).all()


def test_one_hot_from_labels_skip(rng):
    labels = 2*rng.randint(4, size=1000)

    yref = np.zeros((len(labels), labels.max()+1))
    yref[np.arange(len(labels)), labels] = 1
    y = one_hot_from_labels(labels)
    assert np.array_equal(y, yref)


def test_one_hot_from_labels_float(rng):
    classes = rng.uniform(0, 9, size=11)
    inds = rng.randint(len(classes), size=1000)
    labels = classes[inds]

    yref = np.zeros((len(labels), len(classes)))
    yref[np.arange(len(labels)), inds] = 1

    y = one_hot_from_labels(labels, classes=classes)
    assert np.array_equal(y, yref)


def test_spasafe_name():
    assert spasafe_name('UPPER') == 'UPPER'
    assert spasafe_name('Camel') == 'Camel'
    assert spasafe_name('lower') == 'Lower'
    assert spasafe_name('Under_score') == 'Under_score'
    assert spasafe_name(
        'Weird.,:[]^!<>=&\'"symbols', pre_comma_only=False) == 'Weird_symbols'
    assert spasafe_name(
        'Weird.,:[]^!<>=&\'"symbols', pre_comma_only=True) == 'Weird'
    assert spasafe_name('  \tWhite- \t\r\nspace  \r\n') == 'White_space'
    assert spasafe_name('multiple correction\'s') == 'Multiple_corrections'
    assert spasafe_name('123four') == 'Four'
    assert spasafe_name('A') == 'A'
    with pytest.raises(ValueError):
        spasafe_name('')


def test_spasafe_names():
    assert spasafe_names(['A,B', 'A', 'c\'s']) == ['A0', 'A1', 'Cs']



import numpy as np

import nengo
from nengo.utils.builder import remove_passthrough_nodes

from nengo_extras.graphviz import net_diagram, obj_conn_diagram


def make_net():
    with nengo.Network() as model:
        D = 3
        input = nengo.Node([1] * D, label='input')
        a = nengo.networks.EnsembleArray(50, D, label='a')
        b = nengo.networks.EnsembleArray(50, D, label='b')
        output = nengo.Node(None, size_in=D, label='output')

        nengo.Connection(input, a.input, synapse=0.01)
        nengo.Connection(a.output, b.input, synapse=0.01)
        nengo.Connection(b.output, b.input, synapse=0.01, transform=0.9)
        nengo.Connection(a.output, a.input, synapse=0.01,
                         transform=np.ones((D, D)))
        nengo.Connection(b.output, output, synapse=0.01)
    return model


def test_net_diagram():
    """Constructing a .dot file for a network."""
    dot = net_diagram(make_net())
    assert len(dot.splitlines()) == 31


def test_obj_conn_diagram():
    """Constructing a .dot file for a set of objects and connections."""
    model = make_net()
    objs = model.all_nodes + model.all_ensembles
    conns = model.all_connections

    dot = obj_conn_diagram(objs, conns)
    assert len(dot.splitlines()) == 31

    dot = obj_conn_diagram(*remove_passthrough_nodes(objs, conns))
    assert len(dot.splitlines()) == 27



import warnings

import numpy as np

from nengo.builder import Builder, Signal
from nengo.builder.connection import get_eval_points, solve_for_decoders
from nengo.builder.operator import (
    DotInc, ElementwiseInc, Operator, Reset, SimPyFunc)
from nengo.exceptions import ValidationError
from nengo.learning_rules import LearningRuleType
from nengo.params import EnumParam, FunctionParam, NumberParam
from nengo.synapses import Lowpass


class AML(LearningRuleType):
    r"""Association matrix learning rule (AML).

    Enables one-shot learning without catastrophic forgetting of outer product
    association matrices.

    The cue is provided by the pre-synaptic ensemble. The error signal is split
    up: ``error[0]`` provides a scaling factor to the learning rate.
    ``error[1]`` provides a decay rate (i.e., weights are multiplied with this
    value in every time step), ``error[2:]`` provides the target vector.

    The update is given by::

        decoders[...] *= error[1]  # decay
        decoders[...] += alpha * error[0] * error[2:, None] * np.dot(
            pre, base_decoders.T)

    where *alpha* is the learning rate adjusted for *dt* and *base_decoders*
    is the decoder matrix for decoding the identity from the pre-ensemble.

    Parameters
    ----------
    d : int
        Dimensionality of input and output vectors (error signal will be
        *d+2*).
    learning_rate : float, optional
        Learning rate (increase of dot product similarity per second).
    """
    error_type = 'decoded'
    modifies = 'decoders'

    def __init__(self, d, learning_rate=1.):
        super(AML, self).__init__(learning_rate, size_in=d + 2)


class SimAML(Operator):
    def __init__(self, learning_rate, base_decoders, pre, error, decoders,
                 tag=None):
        super(SimAML, self).__init__(tag=tag)

        self.learning_rate = learning_rate
        self.base_decoders = base_decoders

        self.sets = []
        self.incs = []
        self.reads = [pre, error]
        self.updates = [decoders]

    def make_step(self, signals, dt, rng):
        base_decoders = self.base_decoders
        pre = signals[self.pre]
        error = signals[self.error]
        decoders = signals[self.decoders]
        alpha = self.learning_rate * dt

        def step_assoc_learning():
            scale = error[0]
            decay = error[1]
            target = error[2:]
            decoders[...] *= decay
            decoders[...] += alpha * scale * target[:, None] * np.dot(
                pre, base_decoders.T)

        return step_assoc_learning

    @property
    def pre(self):
        return self.reads[0]

    @property
    def error(self):
        return self.reads[1]

    @property
    def decoders(self):
        return self.updates[0]


@Builder.register(AML)
def build_aml(model, aml, rule):
    conn = rule.connection
    rng = np.random.RandomState(model.seeds[conn])

    error = Signal(np.zeros(rule.size_in), name="aml:error")
    model.add_op(Reset(error))
    model.sig[rule]['in'] = error

    pre = model.sig[conn.pre_obj]['in']
    decoders = model.sig[conn]['weights']

    encoders = model.params[conn.pre_obj].encoders
    gain = model.params[conn.pre_obj].gain
    bias = model.params[conn.pre_obj].bias

    eval_points = get_eval_points(model, conn, rng)
    targets = eval_points

    x = np.dot(eval_points, encoders.T)

    wrapped_solver = (model.decoder_cache.wrap_solver(solve_for_decoders)
                      if model.seeded[conn] else solve_for_decoders)
    base_decoders, _ = wrapped_solver(conn, gain, bias, x, targets, rng=rng)

    model.add_op(SimAML(
        aml.learning_rate, base_decoders, pre, error, decoders))


class DeltaRuleFunctionParam(FunctionParam):
    function_test_size = 8  # arbitrary size to test function

    def function_args(self, instance, function):
        return (np.zeros(self.function_test_size),)

    def coerce(self, instance, function):
        function_info = super(DeltaRuleFunctionParam, self).coerce(
            instance, function)

        function, size = function_info
        if function is not None and size != self.function_test_size:
            raise ValidationError(
                "Function '%s' input and output sizes must be equal" %
                function, attr=self.name, obj=instance)

        return function_info


class DeltaRule(LearningRuleType):
    r"""Implementation of the Delta rule.

    By default, this implementation pretends the neurons are linear, and thus
    does not require the derivative of the postsynaptic neuron activation
    function. The derivative function, or a surrogate function, for the
    postsynaptic neurons can be provided in ``post_fn``.

    The update is given by:

        \delta W_ij = \eta a_j e_i f(u_i)

    where ``e_i`` is the input error in the postsynaptic neuron space,
    ``a_j`` is the output activity for presynaptic neuron j,
    ``u_i`` is the input for postsynaptic neuron i,
    and ``f`` is a provided function.

    Parameters
    ----------
    learning_rate : float
        A scalar indicating the rate at which weights will be adjusted.
    pre_tau : float
        Filter constant on the presynaptic output ``a_j``.
    post_fn : callable
        Function ``f`` to apply to the postsynaptic inputs ``u_i``. The
        default of ``None`` means the ``f(u_i)`` term is omitted.
    post_tau : float
        Filter constant on the postsynaptic input ``u_i``. This defaults to
        ``None`` because these should typically be filtered by the connection.
    """
    modifies = 'weights'
    probeable = ('delta', 'in', 'error', 'correction', 'pre', 'post')

    pre_tau = NumberParam('pre_tau', low=0, low_open=True)
    post_tau = NumberParam('post_tau', low=0, low_open=True, optional=True)
    post_fn = DeltaRuleFunctionParam('post_fn', optional=True)
    post_target = EnumParam('post_target', values=('in', 'out'))

    def __init__(self, learning_rate=1e-4, pre_tau=0.005,
                 post_fn=None, post_tau=None, post_target='in'):
        if learning_rate >= 1.0:
            warnings.warn("This learning rate is very high, and can result "
                          "in floating point errors from too much current.")
        self.pre_tau = pre_tau
        self.post_tau = post_tau
        self.post_fn = post_fn
        self.post_target = post_target
        super(DeltaRule, self).__init__(learning_rate, size_in='post')

    @property
    def _argreprs(self):
        args = []
        if self.learning_rate != 1e-4:
            args.append("learning_rate=%g" % self.learning_rate)
        if self.pre_tau != 0.005:
            args.append("pre_tau=%f" % self.pre_tau)
        if self.post_fn is not None:
            args.append("post_fn=%s" % self.post_fn.function)
        if self.post_tau is not None:
            args.append("post_tau=%f" % self.post_tau)
        if self.post_target != 'in':
            args.append("post_target=%s" % self.post_target)

        return args


@Builder.register(DeltaRule)
def build_delta_rule(model, delta_rule, rule):
    conn = rule.connection

    # Create input error signal
    error = Signal(np.zeros(rule.size_in), name="DeltaRule:error")
    model.add_op(Reset(error))
    model.sig[rule]['in'] = error  # error connection will attach here

    # Multiply by post_fn output if necessary
    post_fn = delta_rule.post_fn.function
    post_tau = delta_rule.post_tau
    post_target = delta_rule.post_target
    if post_fn is not None:
        post_sig = model.sig[conn.post_obj][post_target]
        post_synapse = Lowpass(post_tau) if post_tau is not None else None
        post_input = (post_sig if post_synapse is None else
                      model.build(post_synapse, post_sig))

        post = Signal(np.zeros(post_input.shape), name="DeltaRule:post")
        model.add_op(SimPyFunc(post, post_fn, t=None, x=post_input,
                               tag="DeltaRule:post_fn"))
        model.sig[rule]['post'] = post

        error0 = error
        error = Signal(np.zeros(rule.size_in), name="DeltaRule:post_error")
        model.add_op(Reset(error))
        model.add_op(ElementwiseInc(error0, post, error))

    # Compute: correction = -learning_rate * dt * error
    correction = Signal(np.zeros(error.shape), name="DeltaRule:correction")
    model.add_op(Reset(correction))
    lr_sig = Signal(-delta_rule.learning_rate * model.dt,
                    name="DeltaRule:learning_rate")
    model.add_op(DotInc(lr_sig, error, correction, tag="DeltaRule:correct"))

    # delta_ij = correction_i * pre_j
    pre_synapse = Lowpass(delta_rule.pre_tau)
    pre = model.build(pre_synapse, model.sig[conn.pre_obj]['out'])

    model.add_op(Reset(model.sig[rule]['delta']))
    model.add_op(ElementwiseInc(
        correction.column(), pre.row(), model.sig[rule]['delta'],
        tag="DeltaRule:Inc Delta"))

    # expose these for probes
    model.sig[rule]['error'] = error
    model.sig[rule]['correction'] = correction
    model.sig[rule]['pre'] = pre



import mako
import nengo_ocl
import numpy as np
import pyopencl as cl


def plan_aml_decode(queue, pre, base_decoders, decoded, tag=None):
    assert pre.ctype == base_decoders.ctype == decoded.ctype
    assert len(pre) == len(base_decoders) == len(decoded)
    assert np.all(pre.shape0s == base_decoders.shape1s)
    assert np.all(base_decoders.shape0s == decoded.shape0s)

    text = '''
    __kernel void aml_decode(
        __global const int *ds,
        __global const int *ns,
        __global const int *pre_stride0s,
        __global const int *pre_starts,
        __global const ${type} *pre_data,
        __global const int *base_decoders_stride0s,
        __global const int *base_decoders_starts,
        __global const ${type} *base_decoders_data,
        __global const int *decoded_stride0s,
        __global const int *decoded_starts,
        __global ${type} *decoded_data
    ) {
        const int i = get_global_id(0);
        const int k = get_global_id(1);

        const int d = ds[k];
        const int n = ns[k];

        __global const ${type} *pre = pre_data + pre_starts[k];
        __global const ${type} *base_decoders = base_decoders_data +
            base_decoders_starts[k];
        __global ${type} *decoded = decoded_data + decoded_starts[k];

        if (i < n) {
            ${type} x = 0.;
            for (int s = 0; s < d; ++s) {
                x += base_decoders[i * base_decoders_stride0s[k] + s] * pre[s];
            }
            decoded[i] = x;
        }
    }
    '''

    textconf = dict(type=pre.ctype)
    text = nengo_ocl.utils.as_ascii(mako.template.Template(
        text, output_encoding='ascii').render(**textconf))

    full_args = (
        base_decoders.cl_shape1s, base_decoders.cl_shape0s,
        pre.cl_stride0s, pre.cl_starts, pre.cl_buf,
        base_decoders.cl_stride0s, base_decoders.cl_starts,
        base_decoders.cl_buf,
        decoded.cl_stride0s, decoded.cl_starts, decoded.cl_buf,
    )
    _fn = cl.Program(queue.context, text).build().aml_decode
    _fn.set_args(*(arr.data for arr in full_args))

    lsize = None
    gsize = (base_decoders.shape0s.max(), len(pre))
    plan = nengo_ocl.plan.Plan(
        queue, _fn, gsize, lsize=lsize, name="cl_aml_decode", tag=tag)
    plan.full_args = full_args  # prevent garbage collection
    plan.flops_per_call = np.sum(
        base_decoders.shape0s * base_decoders.shape1s * 2 +
        base_decoders.shape1s * 2)
    plan.bw_per_call = decoded.nbytes + pre.nbytes + base_decoders.nbytes

    return plan


def plan_aml(queue, error, decoders, alpha, decoded, tag=None):
    assert error.ctype == decoders.ctype == alpha.ctype == decoded.ctype
    assert len(error) == len(decoders) == len(alpha) == len(decoded)
    assert np.all(error.shape0s - 2 == decoders.shape0s)

    text = '''
    __kernel void aml(
        __global const int *ds,
        __global const int *ns,
        __global const int *error_stride0s,
        __global const int *error_starts,
        __global const ${type} *error_data,
        __global const int *decoders_stride0s,
        __global const int *decoders_starts,
        __global ${type} *decoders_data,
        __global const int *decoded_stride0s,
        __global const int *decoded_starts,
        __global const ${type} *decoded_data,
        __global const ${type} *alphas
    ) {
        const int ij = get_global_id(0);
        const int k = get_global_id(1);

        const int d = ds[k];
        const int n = ns[k];
        const int i = ij / n;
        const int j = ij % n;

        __global ${type} *decoders = decoders_data + decoders_starts[k];
        const ${type} scale = error_data[error_starts[k]];
        const ${type} decay = error_data[error_starts[k] + 1];
        const ${type} error = error_data[error_starts[k] + i + 2];
        const ${type} decoded = decoded_data[decoded_starts[k] + j];
        const ${type} alpha = alphas[k];

        if (i < d) {
            decoders[i * decoders_stride0s[k] + j] *= decay;
            decoders[i * decoders_stride0s[k] + j] += alpha * scale * error *
                decoded;
        }
    }
    '''

    textconf = dict(type=error.ctype)
    text = nengo_ocl.utils.as_ascii(mako.template.Template(
        text, output_encoding='ascii').render(**textconf))

    full_args = (
        decoders.cl_shape0s, decoders.cl_shape1s,
        error.cl_stride0s, error.cl_starts, error.cl_buf,
        decoders.cl_stride0s, decoders.cl_starts, decoders.cl_buf,
        decoded.cl_stride0s, decoded.cl_starts, decoded.cl_buf,
        alpha,
    )
    _fn = cl.Program(queue.context, text).build().aml
    _fn.set_args(*(arr.data for arr in full_args))

    lsize = None
    gsize = (decoders.sizes.max(), len(error))
    plan = nengo_ocl.plan.Plan(
        queue, _fn, gsize, lsize=lsize, name="cl_aml", tag=tag)
    plan.full_args = full_args  # prevent garbage collection
    plan.flops_per_call = np.sum(2 * (error.shape0s * decoded.shape0s))
    plan.bw_per_call = (
        decoded.nbytes + error.nbytes + alpha.nbytes + decoders.nbytes)

    return plan


class AmlSimulator(nengo_ocl.Simulator):
    def plan_SimAML(self, ops):
        alpha = self.Array([op.learning_rate * self.model.dt for op in ops])
        base_decoders = self.RaggedArray(
            [op.base_decoders for op in ops], dtype=np.float32)
        pre = self.all_data[[self.sidx[op.pre] for op in ops]]
        error = self.all_data[[self.sidx[op.error] for op in ops]]
        decoders = self.all_data[[self.sidx[op.decoders] for op in ops]]
        decoded = self.RaggedArray(
            [np.zeros(op.decoders.shape[1]) for op in ops], dtype=np.float32)
        return [
            plan_aml_decode(self.queue, pre, base_decoders, decoded),
            plan_aml(self.queue, error, decoders, alpha, decoded)]



import numpy as np
import pytest

from nengo.utils.compat import range
from nengo.utils.numpy import array_hash

from nengo_extras.spa.utils import circconv, cyclic_vector


@pytest.mark.parametrize('d, k', [(5, 3), (16, 4), (128, 13)])
def test_cyclic_vector(d, k, rng):
    n = 10
    a = rng.normal(scale=1./np.sqrt(d), size=(n, d))
    b = cyclic_vector(d, k, n=n)
    c = circconv(a, b, k=k)
    assert np.allclose(a, c, atol=1e-7), (a, c)


@pytest.mark.parametrize('d, k', [(7, 5), (8, 4)])
def test_cyclic_vector_dist(d, k, rng):
    n = 1000
    v = cyclic_vector(d, k, n=n, rng=rng)
    v = np.round(v, decimals=8)
    v[np.abs(v) < 1e-8] = 0

    hashmap = {}
    counts = {}
    for i in range(n):
        vi = v[i]
        hi = array_hash(vi)
        hashmap.setdefault(hi, vi)
        if hi not in counts:
            counts[hi] = 1
        else:
            counts[hi] += 1

    # check that all generated values are cyclic
    for b in hashmap.values():
        a = rng.normal(scale=1./np.sqrt(d), size=d)
        c = circconv(a, b, k=k)
        assert np.allclose(a, c, atol=1e-7), (a, c)

    # check that the std. dev. of the counts is roughly a Binomial dist.
    values = np.array(list(counts.values()))
    p = 1. / len(values)
    std = np.sqrt(n * p * (1 - p))
    assert np.allclose(values.std(), std, rtol=0.3, atol=0)


def test_cyclic_vector_errors(rng):
    with pytest.raises(ValueError):
        cyclic_vector(2, 1, rng=rng)
    with pytest.raises(ValueError):
        cyclic_vector(1, 2, rng=rng)
    with pytest.raises(ValueError):
        cyclic_vector(2, 2, rng=rng)

    a = cyclic_vector(3, 2, rng=rng)
    assert np.allclose(circconv(a, a), [1, 0, 0])



import nengo
import numpy as np

import nengo_extras


def test_matrix_mult(Simulator, seed, rng, nl, plt):
    shape_left = (2, 2)
    shape_right = (2, 2)

    left_mat = rng.uniform(-1, 1, size=shape_left)
    right_mat = rng.uniform(-1, 1, size=shape_right)

    with nengo.Network("Matrix multiplication test", seed=seed) as model:
        node_left = nengo.Node(left_mat.ravel())
        node_right = nengo.Node(right_mat.ravel())

        with nengo.Config(nengo.Ensemble) as cfg:
            cfg[nengo.Ensemble].neuron_type = nl()
            mult_net = nengo_extras.networks.MatrixMult(
                200, shape_left, shape_right)

        p = nengo.Probe(mult_net.output, synapse=0.01)

        nengo.Connection(node_left, mult_net.input_left)
        nengo.Connection(node_right, mult_net.input_right)

    dt = 0.001
    with Simulator(model, dt=dt) as sim:
        sim.run(1)

    t = sim.trange()
    plt.plot(t, sim.data[p])
    for d in np.dot(left_mat, right_mat).flatten():
        plt.axhline(d, color='k')

    atol, rtol = .15, .01
    ideal = np.dot(left_mat, right_mat).ravel()
    assert np.allclose(sim.data[p][-1], ideal, atol=atol, rtol=rtol)



#!/usr/bin/env python
import imp
import io
import os
import sys

try:
    from setuptools import find_packages, setup
except ImportError:
    raise ImportError(
        "'setuptools' is required but not installed. To install it, "
        "follow the instructions at "
        "https://pip.pypa.io/en/stable/installing/#installing-with-get-pip-py")


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


root = os.path.dirname(os.path.realpath(__file__))
version_module = imp.load_source(
    "version", os.path.join(root, "nengo_extras", "version.py"))
testing = "test" in sys.argv or "pytest" in sys.argv

install_requires = [
    "numpy>=1.8",
    "nengo",
]
plots_require = ["matplotlib>=1.4"]
deepnetworks_require = [
    "keras",
    "scipy",
    "theano",
]
docs_require = [
    "sphinx<1.7",
    "numpydoc>=0.6",
    "guzzle_sphinx_theme",
    "nbsphinx",
    "nbsphinx_link",
    "pillow",
    "jupyter",
    "matplotlib>=1.4",
] + deepnetworks_require + plots_require
tests_require = [
    "jupyter",
    "matplotlib>=1.4",
    "pytest>=3.2",
]


setup(
    name="nengo-extras",
    version=version_module.version,
    author="Applied Brain Research",
    author_email="info@appliedbrainresearch.com",
    packages=find_packages(),
    scripts=[],
    url="https://github.com/nengo/nengo-extras",
    license="Free for non-commercial use",
    description="Lesser used features for the Nengo neural simulator",
    long_description=read("README.rst", "CHANGES.rst"),
    # Without this, `setup.py install` fails to install NumPy.
    # See https://github.com/nengo/nengo/issues/508 for details.
    setup_requires=(["pytest-runner"] if testing else []) + ["numpy>=1.8"],
    install_requires=install_requires,
    extras_require={
        "deepnetworks": deepnetworks_require,
        "docs": docs_require,
        "plots": plots_require,
    },
    tests_require=tests_require,
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Science/Research",
        "License :: Free for non-commercial use",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import numpy as np


def net_diagram(net):
    """Create a .dot file showing nodes, ensmbles, and connections.

    This can be useful for debugging and testing builders that manipulate
    the model graph before construction.

    Parameters
    ----------
    net : Network
        A network from which objects and connections will be extracted.

    Returns
    -------
    text : string
        Text content of the desired .dot file.
    """
    objs = net.all_nodes + net.all_ensembles
    return obj_conn_diagram(objs, net.all_connections)


def obj_conn_diagram(objs, connections):
    """Create a .dot file showing nodes, ensmbles, and connections.

    This can be useful for debugging and testing builders that manipulate
    the model graph before construction.

    Parameters
    ----------
    objs : list of Nodes and Ensembles
        All the nodes and ensembles in the model.
    connections : list of Connections
        All the connections in the model.

    Returns
    -------
    text : string
        Text content of the desired .dot file.
    """
    text = []
    text.append('digraph G {')
    for obj in objs:
        text.append('  "%d" [label="%s"];' % (id(obj), obj.label))

    def label(transform):
        # determine the label for a connection based on its transform
        transform = np.asarray(transform)
        if len(transform.shape) == 0:
            return ''
        return '%dx%d' % transform.shape

    for c in connections:
        text.append('  "%d" -> "%d" [label="%s"];' % (
            id(c.pre_obj), id(c.post_obj), label(c.transform)))
    text.append('}')
    return '\n'.join(text)



import gzip
import io
import os
import re
import tarfile

import nengo
from nengo.utils.compat import is_integer, is_iterable
import numpy as np

from .compat import pickle_load_bytes, urlretrieve


data_dir = nengo.rc.get('nengo_extras', 'data_dir')


def get_file(filename, url):
    filename = os.path.expanduser(filename)
    if not os.path.exists(filename):
        print("Retrieving %r" % url)
        urlretrieve(url, filename=filename)
        print("Data retrieved as %r" % filename)
    return filename


def get_cifar10_tar_gz():
    filename = os.path.join(data_dir, 'cifar-10-python.tar.gz')
    url = 'https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz'
    return get_file(filename, url)


def get_cifar100_tar_gz():
    filename = os.path.join(data_dir, 'cifar-100-python.tar.gz')
    url = 'https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz'
    return get_file(filename, url)


def get_ilsvrc2012_tar_gz():
    filename = os.path.join(data_dir, 'ilsvrc-2012-batches-test3.tar.gz')
    url = 'http://files.figshare.com/5370887/ilsvrc-2012-batches-test3.tar.gz'
    return get_file(filename, url)


def get_mnist_pkl_gz():
    filename = os.path.join(data_dir, 'mnist.pkl.gz')
    url = 'http://deeplearning.net/data/mnist/mnist.pkl.gz'
    return get_file(filename, url)


def get_svhn_tar_gz():
    filename = os.path.join(data_dir, 'svhn-py-colmajor.tar.gz')
    url = 'https://files.figshare.com/7868377/svhn-py-colmajor.tar.gz'
    return get_file(filename, url)


def unpickle_tarfile(tar, name):
    tarextract = tar.extractfile(name)
    return pickle_load_bytes(tarextract)


def load_cifar10(filepath=None, n_train=5, n_test=1, label_names=False):
    """Load the CIFAR-10 dataset.

    Parameters
    ----------
    filepath : str (optional, Default: None)
        Path to the previously downloaded 'cifar-10-python.tar.gz' file.
        If ``None``, the file will be downloaded to the current directory.
    n_train : int (optional, Default: 6)
        The number of training batches to load (max: 6).
    n_test : int (optional, Default: 6)
        The number of testing batches to load (max: 1).
    label_names : boolean (optional, Default: False)
        Whether to provide the category label names.

    Returns
    -------
    train_set : (n_train, n_pixels) ndarray, (n_train,) ndarray
        A tuple of the training image array and label array.
    test_set : (n_test, n_pixels) ndarray, (n_test,) ndarray
        A tuple of the testing image array and label array.
    label_names : list
        A list of the label names.
    """
    if filepath is None:
        filepath = get_cifar10_tar_gz()

    # helper for reading each batch file
    def read_tar_batch(tar, name):
        data = unpickle_tarfile(tar, name)
        return data[b'data'], np.array(data[b'labels'])

    filepath = os.path.expanduser(filepath)
    with tarfile.open(filepath, 'r:gz') as tar:
        if n_train < 1:
            train = (np.array([]), np.array([]))
        else:
            train = ([], [])
            for i in range(n_train):
                data, labels = read_tar_batch(
                    tar, 'cifar-10-batches-py/data_batch_%d' % (i+1))
                train[0].append(data)
                train[1].append(labels)

            train = (np.vstack(train[0]), np.hstack(train[1]))

        if n_test < 1:
            test = (np.array([]), np.array([]))
        else:
            test = read_tar_batch(tar, 'cifar-10-batches-py/test_batch')

        if label_names:
            meta = unpickle_tarfile(tar, 'cifar-10-batches-py/batches.meta')
            names = meta[b'label_names']

    return (train, test) + ((names,) if label_names else ())


def load_cifar100(filepath=None, fine_labels=True, label_names=False):
    """Load the CIFAR-100 dataset.

    Parameters
    ----------
    filepath : str (optional, Default: None)
        Path to the previously downloaded 'cifar-100-python.tar.gz' file.
        If ``None``, the file will be downloaded to the current directory.
    fine_labels : boolean (optional, Default: True)
        Whether to provide the fine labels or coarse labels.
    label_names : boolean (optional, Default: False)
        Whether to provide the category label names.

    Returns
    -------
    train_set : (n_train, n_pixels) ndarray, (n_train,) ndarray
        A tuple of the training image array and label array.
    test_set : (n_test, n_pixels) ndarray, (n_test,) ndarray
        A tuple of the testing image array and label array.
    label_names : list
        A list of the label names.
    """
    if filepath is None:
        filepath = get_cifar100_tar_gz()

    # helper for reading each batch file
    def read_tar_batch(tar, name):
        data = unpickle_tarfile(tar, name)
        return data[b'data'], np.array(
            data[b'fine_labels' if fine_labels else b'coarse_labels'])

    filepath = os.path.expanduser(filepath)
    with tarfile.open(filepath, 'r:gz') as tar:
        train = read_tar_batch(tar, 'cifar-100-python/train')
        test = read_tar_batch(tar, 'cifar-100-python/test')
        if label_names:
            meta = unpickle_tarfile(tar, 'cifar-100-python/meta')
            names = meta[
                b'fine_label_names' if fine_labels else b'coarse_label_names']

    return (train, test) + ((names,) if label_names else ())


def load_ilsvrc2012(filepath=None, n_files=None):
    """Load part of the ILSVRC 2012 (ImageNet) dataset.

    This loads a small section of the ImageNet Large Scale Visual Recognition
    Challenge (ILSVRC) 2012 dataset. The images are from the test portion of
    the dataset, and can be used to test pretrained classifiers.

    Parameters
    ----------
    filepath : str (optional, Default: None)
        Path to the previously downloaded 'ilsvrc-2012-batches-test3.tar.gz'.
        If ``None``, the file will be downloaded to the current directory.
    n_files : int (optional, Default: None)
        Number of files (batches) to load from the archive. Defaults to all.

    Returns
    -------
    images : (n_images, nc, ny, nx) ndarray
        The loaded images. nc = number of channels, ny = height, nx = width
    labels : (n_images,) ndarray
        The labels of the images.
    data_mean : (nc, ny, nx) ndarray
        The mean of the images in the whole of the training set.
    label_names : list
        A list of the label names.
    """
    import PIL.Image  # ``pip install pillow``

    if filepath is None:
        filepath = get_ilsvrc2012_tar_gz()

    # helper for reading each batch file
    def read_tar_batch(tar, name):
        data = unpickle_tarfile(tar, name)
        return data[b'data'], data[b'labels']  # JPEG strings, labels

    def bytes_to_array(b):
        image = PIL.Image.open(io.BytesIO(b))
        array = np.array(image, dtype=np.uint8).reshape(
            image.size[0], image.size[1], 3)
        array = np.transpose(array, (2, 0, 1))
        return array

    filepath = os.path.expanduser(filepath)
    with tarfile.open(filepath, 'r:gz') as tar:
        names = tar.getnames()
        regex = re.compile(r'.*/data_batch_([0-9]+\.[0-9]+)')
        matches = [regex.match(name) for name in names]
        matches = [match for match in matches if match]

        batchfiles = {}
        for match in matches:
            batchfiles[float(match.groups()[-1])] = match.group()

        raw_images = []
        raw_labels = []
        for key in sorted(list(batchfiles))[:n_files]:
            batchfile = batchfiles[key]
            x, y = read_tar_batch(tar, batchfile)
            raw_images.extend(x)
            raw_labels.extend(y)

        n_images = len(raw_images)
        image_shape = bytes_to_array(raw_images[0]).shape
        images = np.zeros((n_images,) + image_shape, dtype=np.uint8)
        for i, s in enumerate(raw_images):
            images[i] = bytes_to_array(s)

        labels = np.array(raw_labels)
        labels.shape = (n_images,)

        meta = unpickle_tarfile(tar, 'batches.meta')
        data_mean = meta[b'data_mean'].reshape(image_shape)
        label_names = meta[b'label_names']

    return images, labels, data_mean, label_names


def load_ilsvrc2012_metadata(filepath=None):
    if filepath is None:
        filepath = get_ilsvrc2012_tar_gz()

    filepath = os.path.expanduser(filepath)
    with tarfile.open(filepath, 'r:gz') as tar:
        meta = unpickle_tarfile(tar, 'batches.meta')
        data_mean = meta[b'data_mean'].reshape((3, 256, 256))
        label_names = meta[b'label_names']

    return data_mean, label_names


def load_mnist(filepath=None, validation=False):
    """Load the MNIST dataset.

    Parameters
    ----------
    filepath : str (optional, Default: None)
        Path to the previously downloaded 'mnist.pkl.gz' file.
        If ``None``, the file will be downloaded to the current directory.
    validation : boolean (optional, Default: False)
        Whether to provide the validation data as a separate set (True),
        or combine it into the training data (False).

    Returns
    -------
    train_set : (n_train, n_pixels) ndarray, (n_train,) ndarray
        A tuple of the training image array and label array.
    validation_set : (n_valid, n_pixels) ndarray, (n_valid,) ndarray
        A tuple of the validation image array and label array
        (if ``validation``)
    test_set : (n_test, n_pixels) ndarray, (n_test,) ndarray
        A tuple of the testing image array and label array.
    """
    if filepath is None:
        filepath = get_mnist_pkl_gz()

    filepath = os.path.expanduser(filepath)
    with gzip.open(filepath, 'rb') as f:
        train_set, valid_set, test_set = pickle_load_bytes(f)

    if validation:
        return train_set, valid_set, test_set
    else:  # combine valid into train
        train_set = (np.vstack((train_set[0], valid_set[0])),
                     np.hstack((train_set[1], valid_set[1])))
        return train_set, test_set


def load_svhn(filepath=None, n_train=9, n_test=3, data_mean=False,
              label_names=False):
    """Load the SVHN dataset.

    Parameters
    ----------
    filepath : str (optional, Default: None)
        Path to the previously downloaded 'svhn-py-colmajor.tar.gz' file.
        If ``None``, the file will be downloaded to the current directory.
    n_train : int (optional, Default: 6)
        The number of training batches to load (max: 6).
    n_test : int (optional, Default: 6)
        The number of testing batches to load (max: 1).
    label_names : boolean (optional, Default: False)
        Whether to provide the category label names.

    Returns
    -------
    train_set : (n_train, n_pixels) ndarray, (n_train,) ndarray
        A tuple of the training image array and label array.
    test_set : (n_test, n_pixels) ndarray, (n_test,) ndarray
        A tuple of the testing image array and label array.
    label_names : list
        A list of the label names.
    """
    shape = (3, 32, 32)

    if filepath is None:
        filepath = get_svhn_tar_gz()

    def read_tar_batch(tar, name):
        data = unpickle_tarfile(tar, name)
        return data[b'data'], np.array(data[b'labels'])

    def load_batches(tar, inds):
        if len(inds) < 1:
            return (np.array([]), np.array([]))

        batches = ([], [])
        for i in inds:
            data, labels = read_tar_batch(
                tar, 'svhn-py-colmajor/data_batch_%d' % i)
            batches[0].append(data.T)
            batches[1].append(labels)

        return (np.vstack(batches[0]).reshape((-1,) + shape),
                np.hstack(batches[1]))

    filepath = os.path.expanduser(filepath)
    with tarfile.open(filepath, 'r:gz') as tar:
        train = load_batches(tar, list(range(1, n_train+1)))
        test = load_batches(tar, list(range(10, n_test+10)))

        if label_names or data_mean:
            meta = unpickle_tarfile(tar, 'svhn-py-colmajor/batches.meta')
        data_mean = (meta[b'data_mean'].reshape(shape),) if data_mean else ()
        label_names = (meta[b'label_names'],) if label_names else ()

    return (train, test) + data_mean + label_names


def spasafe_name(name, pre_comma_only=True):
    """Make a name safe to use as a SPA semantic pointer name.

    Ensure a name conforms with SPA name standards. Replaces hyphens and
    spaces with underscores, removes all other characters, and makes the
    first letter uppercase.

    Parameters
    ----------
    pre_comma_only : boolean
        Only use the part of a name before a/the first comma.
    """
    if len(name) == 0:
        raise ValueError("Empty name.")

    if pre_comma_only and ',' in name:
        name = name.split(',')[0]  # part before first comma
    name = name.strip()
    name = re.sub(r'(\s|-|,)+', '_', name)  # repl space/hyphen/comma w undersc
    name = re.sub('(^[^a-zA-Z]+)|[^a-zA-Z0-9_]+', '', name)  # del other chars
    name = name[0].upper() + name[1:]  # capitalize first letter
    return name


def spasafe_names(label_names, pre_comma_only=True):
    """Make names safe to use as SPA semantic pointer names.

    Format a list of names to conform with SPA name standards. In addition
    to running each name through ``spasafe_name``, this function numbers
    duplicate names so they are unique.

    Parameters
    ----------
    pre_comma_only : boolean
        Only use the part of a name before a/the first comma.
    """
    vocab_names = [spasafe_name(name, pre_comma_only=pre_comma_only)
                   for name in label_names]

    # number duplicates
    unique = set()
    duplicates = []
    for name in vocab_names:
        if name in unique:
            duplicates.append(name)
        else:
            unique.add(name)

    duplicates = {name: 0 for name in duplicates}
    for i, name in enumerate(vocab_names):
        if name in duplicates:
            vocab_names[i] = '%s%d' % (name, duplicates[name])
            duplicates[name] += 1

    return vocab_names


def one_hot_from_labels(labels, classes=None, dtype=float):
    """Turn integer labels into a one-hot encoding.

    Parameters
    ==========
    labels : (n,) array
        Labels to turn into one-hot encoding.
    classes : int or (n_classes,) array (optional)
        Classes for encoding. If integer and ``labels.dtype`` is integer, this
        is the number of classes in the encoding. If iterable, this is the
        list of classes to place in the one-hot (must be a superset of the
        unique elements in ``labels``).
    dtype : dtype (optional)
        Data type of returned one-hot encoding (defaults to ``float``).
    """
    assert labels.ndim == 1
    n = labels.shape[0]

    if np.issubdtype(labels.dtype, np.integer) and (
            classes is None or is_integer(classes)):
        index = labels
        index_min, index_max = index.min(), index.max()
        n_classes = (index_max + 1) if classes is None else classes
        assert index_min >= 0
        assert index_max < n_classes
    else:
        if classes is not None:
            assert is_iterable(classes)
            assert set(np.unique(labels)).issubset(classes)
        classes = np.unique(labels) if classes is None else classes
        n_classes = len(classes)

        c_index = np.argsort(classes)
        c_sorted = classes[c_index]
        index = c_index[np.searchsorted(c_sorted, labels)]

    y = np.zeros((n, n_classes), dtype=dtype)
    y[np.arange(n), index] = 1
    return y


class ZCAWhiten(object):
    """ZCA Whitening

    References
    ----------
    .. [1] Krizhevsky, Alex. "Learning multiple layers of features from tiny
           images" (2009) MSc Thesis, Dept. of Comp. Science, Univ. of
           Toronto. pp. 48-49.
    """

    def __init__(self, beta=1e-2, gamma=1e-5):
        self.beta = beta
        self.gamma = gamma

        self.dims = None
        self.pixel_mu = None
        self.e = None
        self.V = None
        self.Sinv = None

    def contrast_normalize(self, X, remove_mean=True, beta=None,
                           hard_beta=True):
        X = np.asarray(X, dtype=np.float64)
        if X.ndim != 2:
            raise ValueError('contrast_normalize requires flat patches')

        Xc = X - X.mean(axis=1)[:, None] if remove_mean else X
        l2 = (Xc * Xc).sum(axis=1)

        beta = self.beta if beta is None else beta
        div2 = np.maximum(l2, beta) if hard_beta else l2 + beta
        return Xc / np.sqrt(div2[:, None])

    def fit(self, X):
        """Fit whitening transform to training data

        Parameters
        ----------
        X : array_like
            Flattened data, with each row corresponding to one example
        """
        X = self.contrast_normalize(X)
        self.dims = X.shape[1]

        self.pixel_mu = X.mean(axis=0)
        X -= self.pixel_mu[None, :]  # each pixel has zero mean

        S = np.dot(X.T, X) / (X.shape[0] - 1)
        e, V = np.linalg.eigh(S)
        self.e = e
        self.V = V

        self.Sinv = np.dot(np.sqrt(1.0 / (e + self.gamma)) * V, V.T)

        return np.dot(X, self.Sinv)

    def transform(self, X):
        assert self.dims is not None

        X = self.contrast_normalize(X, beta=self.beta)
        assert X.shape[1] == self.dims

        X -= self.pixel_mu[None, :]
        return np.dot(X, self.Sinv)



"""Export to GEXF for visualization of networks in Gephi."""

from collections import Mapping, namedtuple, OrderedDict, Sequence
from datetime import date
import weakref
import xml.etree.ElementTree as et

import nengo
try:
    import nengo_spa as spa
except ImportError:
    spa = None
import numpy as np


class DispatchTable(object):
    """A descriptor to dispatch to other methods depending on argument type.

    How to use: assign the descriptor to a class attribute and use the
    ``register`` decorator to declare which functions to dispatch to
    for specific types::

        class MyClass(object):
            dispatch = DispatchTable()

            @dispatch.register(TypeA)
            def handle_type_a(self, obj_of_type_a):
                # ...

            @dispatch.register(TypeB)
            def handle_type_b(self, obj_of_type_b):
                # ...

    To then call the method for the appropriate type::

        inst = MyClass()
        inst.dispatch(obj_of_type_a_or_b)

    If multiple methods would match (e.g. if *TypeB* inherits from *TypeA*),
    the most specific method will be used (to be precise: the first type in the
    method resolution order with a registered method will be used).

    The *DispatchTable* descriptor accepts another *DispatchTable* as argument
    which will be used as a fallback. This allows to inherit the dispatch
    table and selectively overwrite methods like so::

        class Inherited(MyClass):
            dispatch = DispatchTable(MyClass.dispatch)

            @dispatch.register(TypeA)
            def alternate_type_a_handler(self, obj_of_type_a):
                # ...

    Finally, dispatch methods can also be changed on a per-instance basis::

        inst.dispatch.register(TypeA, inst_type_a_handler)
    """

    class InstDispatch(object):
        """Return value when accessing the dispatch table on an instance."""
        __slots__ = ('param', 'inst', 'owner')

        def __init__(self, param, inst, owner):
            self.param = param
            self.inst = inst
            self.owner = owner

        def __call__(self, obj):
            for cls in obj.__class__.__mro__:
                if cls in self.param.inst_type_table.get(self.inst, {}):
                    return self.param.inst_type_table[self.inst][cls](obj)
                elif cls in self.param.type_table:
                    return self.param.type_table[cls](self.inst, obj)
                elif self.param.parent is not None:
                    try:
                        return self.param.parent.__get__(
                            self.inst, self.owner)(obj)
                    except NotImplementedError:
                        pass
            raise NotImplementedError(
                "Nothing to dispatch to for type {}.".format(type(obj)))

        def register(self, type_, fn):
            if self.inst not in self.param.inst_type_table:
                self.param.inst_type_table[self.inst] = (
                    weakref.WeakKeyDictionary())
            table = self.param.inst_type_table[self.inst]
            table[type_] = fn
            return fn

    def __init__(self, parent=None):
        self.type_table = weakref.WeakKeyDictionary()
        self.inst_type_table = weakref.WeakKeyDictionary()
        self.parent = parent

    def register(self, type_):
        def _register(fn):
            assert type_ not in self.type_table
            self.type_table[type_] = fn
            return fn
        return _register

    def __get__(self, inst, owner):
        if inst is None:
            return self
        return self.InstDispatch(self, inst, owner)


class HierarchicalLabeler(object):
    """Obtains labels for objects in a Nengo network.

    The names will include the network hierarchy.

    Usage example::

        labels = HierarchicalLabeler().get_labels(model)
    """

    dispatch = DispatchTable()

    def __init__(self):
        self._names = None

    @dispatch.register(Sequence)
    def get_labels_from_sequence(self, seq):
        base_name = self._names[seq]
        for i, obj in enumerate(seq):
            self._handle_found_name(obj, '{base_name}[{i}]'.format(
                base_name=base_name, i=i))

    @dispatch.register(Mapping)
    def get_labels_from_mapping(self, mapping):
        base_name = self._names[mapping]
        for k in mapping:
            obj = mapping[k]
            self._handle_found_name(obj, '{base_name}[{k}]'.format(
                base_name=base_name, k=k))

    @dispatch.register(object)
    def get_labels_from_object(self, obj):
        pass

    @dispatch.register(nengo.Network)
    def get_labels_from_network(self, net):
        if net in self._names:
            base_name = self._names[net] + '.'
        else:
            base_name = ''

        check_last = {
            'ensembles', 'nodes', 'connections', 'networks', 'probes'}
        check_never = {
            'all_ensembles', 'all_nodes', 'all_connections', 'all_networks',
            'all_objects', 'all_probes'}

        for name in dir(net):
            if (not name.startswith('_') and
                    name not in check_last | check_never):
                try:
                    attr = getattr(net, name)
                except AttributeError:
                    pass
                else:
                    self._handle_found_name(attr, base_name + name)

        for name in check_last:
            attr = getattr(net, name)
            self._handle_found_name(attr, base_name + name)

    def _handle_found_name(self, obj, name):
        if (isinstance(obj, (nengo.base.NengoObject, nengo.Network)) and
                obj not in self._names):
            self._names[obj] = name
            self.dispatch(obj)

    def get_labels(self, model):
        self._names = weakref.WeakKeyDictionary()
        self.dispatch(model)
        return self._names


Attr = namedtuple('Attr', ['id', 'type', 'default'])


class GexfConverter(object):
    """Converts Nengo models into GEXF files.

    This can be loaded in Gephi for visualization of the model graph.

    Links:

    * `Gephi <https://gephi.org/>`_
    * `GEXF <https://github.com/gephi/gexf/wiki>`_

    This class can be inherited from to customize the conversion or
    alternatively the ``dispatch`` table can be changed on a
    per-instance basis.

    Note that probes are currently not included in the graph.

    The following attributes will be stored on graph nodes:

    * *type*: type of the Nengo object (e.g., *nengo.ensemble.Ensemble*),
    * *net*: unique ID of the containing network,
    * *net_label*: (possibly non-unique) label of the containing network,
    * *size_in*: input size,
    * *size_out*: output_size,
    * *radius*: ensemble radius (unset for other nodes),
    * *n_neurons*: number of neurons (0 for non-ensembles),
    * *neuron_type*: string representation of the neuron type (unset for
      non-ensembles).

    The following attributes will be stored on graph edges:

    * *pre_type*: type of the connection's pre object (e.g.,
      *nengo.ensemble.Neurons*),
    * *post_type*: type of the connection's post object (e.g.,
      *nengo.ensemble.Neurons*),
    * *synapse*: string representation of the synapse,
    * *tau*: the tau parameter of the synapse if existent,
    * *function*: string representation of the connection's function,
    * *transform*: string representation of the connection's transform,
    * *scalar_transform*: float representation of the transform if it is a
      scalar,
    * *learning_rule_type*: string representation of the connection's learning
      rule type.

    Parameters
    ----------
    labeler : optional
        Object with a ``get_labels`` method that returns a dictionary mapping
        model objects to labels. If not given, a new `HierarchicalLabeler`
        will be used.
    hierarchical : bool, optional (default: False)
        Whether to include information of the network hierarchy in the file.
        Support for hierarchical graphs was removed in Gephi 0.9 and
        hierarchical networks will be automatically flattened which leaves an
        unconnected node for every network.

    Examples
    --------

    Basic usage to write a GEXF file::

        GexfConverter().convert(model).write('model.gexf')
    """

    dispatch = DispatchTable()

    node_attrs = OrderedDict((
        ('type', Attr(0, 'string', None)),
        ('net', Attr(1, 'long', None)),
        ('net_label', Attr(2, 'string', None)),
        ('size_in', Attr(3, 'integer', None)),
        ('size_out', Attr(4, 'integer', None)),
        ('radius', Attr(5, 'float', None)),
        ('n_neurons', Attr(6, 'integer', 0)),
        ('neuron_type', Attr(7, 'string', None)),
        ))
    edge_attrs = OrderedDict((
        ('pre_type', Attr(0, 'string', None)),
        ('post_type', Attr(1, 'string', None)),
        ('synapse', Attr(2, 'string', None)),
        ('tau', Attr(3, 'float', None)),
        ('function', Attr(4, 'string', None)),
        ('transform', Attr(5, 'string', None)),
        ('scalar_transform', Attr(6, 'float', 1.)),
        ('learning_rule_type', Attr(7, 'string', None)),
    ))

    def __init__(self, labeler=None, hierarchical=False):
        if labeler is None:
            labeler = HierarchicalLabeler()
        self.labeler = labeler
        self.hierarchical = hierarchical
        self.version = (1, 3)
        self.tag = 'draft'

        # State used during processing of a model
        # WeakKeyDict so we don't prevent garbage collection after conversion
        # finished.
        self._labels = weakref.WeakKeyDictionary()
        self._net = None

    def convert(self, model):
        """Convert a model to GEXF format.

        Returns
        -------
        xml.etree.ElementTree.ElementTree
            Converted model.
        """
        self._labels = self.labeler.get_labels(model)
        self._labels[model] = 'model'
        return self.make_document(model)

    def make_document(self, model):
        """Create the GEXF XML document from *model*.

        This method is exposed so it can be overwritten in inheriting classes.
        Invoke `convert` instead of this method to convert a model.

        Returns
        -------
        xml.etree.ElementTree.ElementTree
            Converted model.
        """
        version = '.'.join(str(i) for i in self.version)
        tag_version = version + self.tag
        gexf = et.Element('gexf', {
            'xmlns': 'http://www.gexf.net/' + tag_version,
            'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
            'xsi:schemaLocation': (
                'http://www.gexf.net/' + tag_version + ' ' +
                'http://www.gexf.net/' + tag_version + '/gexf.xsd'),
            'version': version
        })

        meta = et.SubElement(gexf, 'meta', {
            'lastmodifieddate': date.today().isoformat()})
        creator = et.SubElement(meta, 'creator')
        creator.text = self.get_typename(self)

        graph = et.SubElement(gexf, 'graph', {'defaultedgetype': 'directed'})
        graph.append(self.make_attr_defs('node', self.node_attrs))
        graph.append(self.make_attr_defs('edge', self.edge_attrs))

        graph.append(self.dispatch(model))

        edges = et.SubElement(graph, 'edges')
        for c in model.all_connections:
            elem = self.dispatch(c)
            if elem is not None:
                edges.append(elem)

        return et.ElementTree(gexf)

    def make_attr_defs(self, cls, defs):
        """Generate an attribute definition block.

        Parameters
        ----------
        cls : str
            Class the attribute definitions are for ('node' or 'edge').
        defs : dict
            Attribute definitions. Maps attribute names to `Attr` instances.

        Returns
        -------
        xml.etree.ElementTree.Element
        """
        attributes = et.Element('attributes', {'class': cls})
        for k, d in defs.items():
            attr = et.SubElement(attributes, 'attribute', {
                'id': str(d.id),
                'title': k,
                'type': d.type,
            })
            if d.default is not None:
                default = et.SubElement(attr, 'default')
                default.text = str(d.default)
        return attributes

    def make_attrs(self, defs, attrs):
        """Generates a block of attribute values.

        Parameters
        ----------
        defs : dict
            Attribute definitions. Maps attribute names to `Attr` instances.
        attrs : dict
            Mapping of attribute names to assigned values.

        Returns
        -------
        xml.etree.ElementTree.Element
        """
        values = et.Element('attvalues')
        assert all(k in defs for k in attrs.keys())
        for k, d in defs.items():
            if k in attrs and attrs[k] is not None:
                values.append(et.Element('attvalue', {
                    'for': str(d.id),
                    'value': str(attrs[k]),
                }))
        return values

    def make_node(self, obj, **attrs):
        """Generate a node for *obj* with attributes *attrs*."""
        tag_attrib = {'id': str(id(obj))}
        if obj in self._labels:
            tag_attrib['label'] = self._labels[obj]
        node = et.Element('node', tag_attrib)
        if len(attrs) > 0:
            node.append(self.make_attrs(self.node_attrs, attrs))
        return node

    def make_edge(self, obj, source, target, **attrs):
        "Edge for *obj* from *source* to *target* with attributes *attrs*."
        tag_attrib = {
            'id': str(id(obj)),
            'source': str(id(source)),
            'target': str(id(target))
        }
        edge = et.Element('edge', tag_attrib)
        if len(attrs) > 0:
            edge.append(self.make_attrs(self.edge_attrs, attrs))
        return edge

    @dispatch.register(nengo.Network)
    def convert_network(self, net):
        parent_net = self._net
        self._net = net

        nodes = et.Element('nodes')
        leaves = net.ensembles + net.nodes + net.probes
        for leave in leaves:
            leave_elem = self.dispatch(leave)
            if leave_elem is not None:
                nodes.append(leave_elem)
        if self.hierarchical:
            for subnet in net.networks:
                subnet_node = self.make_node(
                    subnet, type=self.get_typename(subnet), net=id(self._net),
                    net_label=self._labels.get(self._net, None),
                    n_neurons=subnet.n_neurons)
                subnet_node.append(self.dispatch(subnet))
                nodes.append(subnet_node)
        else:
            for subnet in net.networks:
                nodes.extend(self.dispatch(subnet))

        self._net = parent_net
        return nodes

    @dispatch.register(nengo.Ensemble)
    def convert_ensemble(self, ens):
        return self.make_node(
            ens,
            type=self.get_typename(ens),
            net=id(self._net),
            net_label=self._labels.get(self._net, None),
            size_in=ens.dimensions,
            size_out=ens.dimensions,
            radius=ens.radius,
            n_neurons=ens.n_neurons,
            neuron_type=ens.neuron_type,
        )

    @dispatch.register(nengo.Node)
    def convert_node(self, node):
        return self.make_node(
            node,
            type=self.get_typename(node),
            net=id(self._net),
            net_label=self._labels.get(self._net, None),
            size_in=node.size_in,
            size_out=node.size_out,
        )

    @dispatch.register(nengo.Probe)
    def convert_probe(self, probe):
        return None

    @dispatch.register(nengo.Connection)
    def convert_connection(self, conn):
        source = self.get_node_obj(conn.pre_obj)
        target = self.get_node_obj(conn.post_obj)
        return self.make_edge(
            conn, source, target,
            pre_type=self.get_typename(conn.pre_obj),
            post_type=self.get_typename(conn.post_obj),
            synapse=conn.synapse,
            tau=conn.synapse.tau if hasattr(conn.synapse, 'tau') else None,
            function=conn.function,
            transform=conn.transform,
            scalar_transform=(
                conn.transform if np.isscalar(conn.transform) else None),
            learning_rule_type=conn.learning_rule_type
        )

    def get_node_obj(self, obj):
        """Get an object with a corresponding graph node related to *obj*.

        For certain objects like `nengo.ensemble.Neurons` or
        `nengo.connection.LearningRule` no graph node will be created. This
        function will resolve such an object to a related object that has a
        corresponding graph node (e.g., the ensemble for a neurons object or
        the pre object for a learning rule).

        In `GexfConverter` this is used to make sure connections are between
        the correct nodes and do not introduce unrelated dangling nodes.
        """
        if isinstance(obj, nengo.ensemble.Neurons):
            return obj.ensemble
        elif isinstance(obj, nengo.connection.LearningRule):
            return self.get_node_obj(obj.connection.pre_obj)
        return obj

    @classmethod
    def get_typename(cls, obj):
        tp = type(obj)
        return tp.__module__ + '.' + tp.__name__


class CollapsingGexfConverter(GexfConverter):
    """Converts Nengo models into GEXF files with some collapsed networks.

    See `GexfConverter` for general information on conversion to GEXF files.
    This class will collapse certain networks to a single node in the
    conversion.

    Parameters
    ----------
    to_collapse : sequence, optional
        Network types to collapse, if not given the networks listed in
        ``NENGO_NETS`` and ``SPA_NETS`` will be collapsed. Note that
        ``SPA_NETS`` currently only contains networks from *nengo_spa*,
        but not the *spa* module in core *nengo*.
    labeler : optional
        Object with a ``get_labels`` method that returns a dictionary mapping
        model objects to labels. If not given, a new `HierarchicalLabeler`
        will be used.
    hierarchical : bool, optional (default: False)
        Whether to include information of the network hierarchy in the file.
        Support for hierarchical graphs was removed in Gephi 0.9 and
        hierarchical networks will be automatically flattened which leaves an
        unconnected node for every network.
    """
    dispatch = DispatchTable(GexfConverter.dispatch)

    NENGO_NETS = (
        nengo.networks.CircularConvolution,
        nengo.networks.EnsembleArray,
        nengo.networks.Product)
    if spa is None:
        SPA_NETS = ()
    else:
        SPA_NETS = (
            spa.networks.CircularConvolution,
            spa.AssociativeMemory,
            spa.Bind,
            spa.Compare,
            spa.Product,
            spa.Scalar,
            spa.State,
            spa.Transcode)

    def __init__(self, to_collapse=None, labeler=None, hierarchical=False):
        super(CollapsingGexfConverter, self).__init__(
            labeler=labeler, hierarchical=hierarchical)

        if to_collapse is None:
            to_collapse = self.NENGO_NETS + self.SPA_NETS

        for cls in to_collapse:
            self.dispatch.register(cls, self.convert_collapsed)

        self.obj2collapsed = weakref.WeakKeyDictionary()

    def convert_collapsed(self, net):
        """Used to convert a network into a collapsed graph node."""
        nodes = et.Element('nodes')
        nodes.append(self.make_node(
            net, type=self.get_typename(net), net=id(self._net),
            net_label=self._labels.get(self._net, None),
            n_neurons=net.n_neurons))
        self.obj2collapsed.update({
            child: net for child in net.all_objects})
        return nodes

    def get_node_obj(self, obj):
        obj = super(CollapsingGexfConverter, self).get_node_obj(obj)
        return self.obj2collapsed.get(obj, obj)



from __future__ import absolute_import

import numpy as np


def circconv(a, b, k=1, invert_a=False, invert_b=False, axis=-1):
    """A reference Numpy implementation of circular convolution.

    If ``k == 1``, returns ``a * b``, where ``*`` represents circular
    convolution.

    If ``k > 1``, return ``a * b * ... * b``, with ``k`` copies of ``b``.

    Parameters
    ----------
    a, b : array_like
        Arrays to convolve.
    k : int (optional)
        Number of times to convolve ``b`` with ``a``. Defaults to 1.
    invert_a : boolean
        Whether to invert ``a``.
    invert_b : boolean
        Whether to invert ``b``.
    axis : int (optional)
        Axis along which to perform the convolution. Defaults to the last axis.
    """
    A = np.fft.fft(a, axis=axis)
    B = np.fft.fft(b, axis=axis)
    if invert_a:
        A = A.conj()
    if invert_b:
        B = B.conj()
    if k != 1:
        B = B**k
    return np.fft.ifft(A * B, axis=axis).real


def _shared_factor_set(k):
    """Returns a set of all numbers sharing a common factor with `k`"""
    f = []
    sqrt = lambda x: int(np.sqrt(x))
    for i in range(2, sqrt(k) + 1):
        if k % i == 0:
            ki = int(k / i)
            f.extend(p * i for p in range(1, ki))
            f.extend(p * ki for p in range(1, i))

    return set(f)


def cyclic_vector(d, k, n=None, rng=np.random):
    """Leaves a target vector unchanged after k convolutions.

    For example, if a is any vector and b = cyclic_vector(d, 4):

        a * b * b * b * b = a

    where * is the circular convolution operator. By corollary:

        b * b * b * b = [1, 0, ..., 0]

    Parameters
    ----------
    d : int
        The number of dimensions of the generated vector(s).
    k : int
        The number of convolutions required to reproduce the input.
    n : int (optional)
        The number of vectors to generate.
    rng : random number generator (optional)
        The random number generator to use.

    Output
    ------
    u : (n, d) array
        Array of vector(s). If 'n' is None, the shape will be `(d,)`.
    """
    d, k = int(d), int(k)
    if k < 2:
        raise ValueError("'k' must be at least 2 (got %d)" % k)
    if d < 3:
        raise ValueError("'d' must be at least 3 (got %d)" % d)

    d2 = (d - 1) // 2
    nn = 1 if n is None else int(n)

    # Pick roots r such that r**k == 1
    roots = np.exp(2.j * np.pi / k * np.arange(k))
    rootpow = rng.randint(0, k, size=(nn, d2))

    # Ensure at least one root power in each vector is coprime with k, so that
    # no vector will take LESS than k convolutions to reproduce itself.
    coprimes = set(range(1, k)) - _shared_factor_set(k)
    for i in range(nn):
        # TODO: better method than rejection sampling?
        while not any(p in coprimes for p in rootpow[i]):
            rootpow[i] = rng.randint(0, k, size=d2)

    # Create array of Fourier coefficients such that U**k == ones(d)
    U = np.ones((nn, d), dtype=np.complex128)
    U[:, 1:1+d2] = roots[rootpow]

    # For even k, U[:, d2+1] = 1 or -1 are both valid
    if k % 2 == 0:
        U[:, 1+d2] = 2 * rng.randint(0, 2, size=nn) - 1

    # Respect Fourier symmetry conditions for real vectors
    U[:, -d2:] = U[:, d2:0:-1].conj()

    u = np.fft.ifft(U, axis=1).real
    return u[0] if n is None else u



import numpy as np

import nengo


def MatrixMult(n_neurons, shape_left, shape_right, net=None):
    """Computes the matrix product A*B.

    Both matrices need to be two dimensional.

    See the `Matrix Multiplication example <:doc:networks>`_
    for a description of the network internals.

    Parameters
    ----------
    n_neurons : int
        Number of neurons used per product of two scalars.

        .. note:: If an odd number of neurons is given, one less neuron will be
                  used per product to obtain an even number. This is due to
                  the implementation the `.Product` network.
    shape_left : tuple
        Shape of the A input matrix.
    shape_right : tuple
        Shape of the B input matrix.
    net : Network, optional (Default: None)
        A network in which the network components will be built.
        This is typically used to provide a custom set of Nengo object
        defaults through modifying ``net.config``.

    Returns
    -------
    net : Network
        The newly built matrix multiplication network, or the provided ``net``.
    """

    if len(shape_left) != 2:
        raise ValueError("Shape {} is not two dimensional.".format(shape_left))
    if len(shape_right) != 2:
        raise ValueError(
            "Shape {} is not two dimensional.".format(shape_right))
    if shape_left[1] != shape_right[0]:
        raise ValueError(
            "Matrix dimensions {} and  {} are incompatible".format(
                shape_left, shape_right))

    if net is None:
        net = nengo.Network(label="Matrix multiplication")

    size_left = np.prod(shape_left)
    size_right = np.prod(shape_right)

    with net:
        net.input_left = nengo.Node(size_in=size_left)
        net.input_right = nengo.Node(size_in=size_right)

        # The C matrix is composed of populations that each contain
        # one element of A (left) and one element of B (right).
        # These elements will be multiplied together in the next step.
        size_c = size_left * shape_right[1]
        net.C = nengo.networks.Product(n_neurons, size_c)

        # Determine the transformation matrices to get the correct pairwise
        # products computed.  This looks a bit like black magic but if
        # you manually try multiplying two matrices together, you can see
        # the underlying pattern.  Basically, we need to build up D1*D2*D3
        # pairs of numbers in C to compute the product of.  If i,j,k are the
        # indexes into the D1*D2*D3 products, we want to compute the product
        # of element (i,j) in A with the element (j,k) in B.  The index in
        # A of (i,j) is j+i*D2 and the index in B of (j,k) is k+j*D3.
        # The index in C is j+k*D2+i*D2*D3, multiplied by 2 since there are
        # two values per ensemble.  We add 1 to the B index so it goes into
        # the second value in the ensemble.
        transform_left = np.zeros((size_c, size_left))
        transform_right = np.zeros((size_c, size_right))

        for i, j, k in np.ndindex(shape_left[0], *shape_right):
            c_index = (j + k * shape_right[0] + i * size_right)
            transform_left[c_index][j + i * shape_right[0]] = 1
            transform_right[c_index][k + j * shape_right[1]] = 1

        nengo.Connection(
            net.input_left, net.C.A, transform=transform_left, synapse=None)
        nengo.Connection(
            net.input_right, net.C.B, transform=transform_right, synapse=None)

        # Now do the appropriate summing
        size_output = shape_left[0] * shape_right[1]
        net.output = nengo.Node(size_in=size_output)

        # The mapping for this transformation is much easier, since we want to
        # combine D2 pairs of elements (we sum D2 products together)
        transform_c = np.zeros((size_output, size_c))
        for i in range(size_c):
            transform_c[i // shape_right[0]][i] = 1

        nengo.Connection(
            net.C.output, net.output, transform=transform_c, synapse=None)

    return net



import numpy as np

import nengo
import nengo.utils.numpy as npext
from nengo.exceptions import ValidationError
from nengo.params import NumberParam


def softplus(x, sigma=1.):
    x = np.asarray(x)
    y = x / sigma
    z = np.array(x)
    z[y < 34.0] = sigma * np.log1p(np.exp(y[y < 34.0]))
    return z
    # ^ 34.0 gives exact answer in 32 or 64 bit but doesn't overflow in 32 bit


def lif_j(j, tau_ref, tau_rc, amplitude=1.):
    return amplitude / (tau_ref + tau_rc * np.log1p(1. / j))


class SoftLIFRate(nengo.neurons.LIFRate):
    """LIF neuron with smoothing around the firing threshold.

    This is a rate version of the LIF neuron whose tuning curve has a
    continuous first derivative, due to the smoothing around the firing
    threshold. It can be used as a substitute for LIF neurons in deep networks
    during training, and then replaced with LIF neurons when running
    the network [1]_.

    Parameters
    ----------
    sigma : float
        Amount of smoothing around the firing threshold. Larger values mean
        more smoothing.
    amplitude : float
        Scaling factor on the output. If 1 (default), output rates correspond
        to LIF neuron model rates.
    tau_rc : float
        Membrane RC time constant, in seconds. Affects how quickly the membrane
        voltage decays to zero in the absence of input (larger = slower decay).
    tau_ref : float
        Absolute refractory period, in seconds. This is how long the
        membrane voltage is held at zero after a spike.

    References
    ----------
    .. [1] E. Hunsberger & C. Eliasmith (2015). Spiking Deep Networks with
       LIF Neurons. arXiv Preprint, 1510. https://arxiv.org/abs/1510.08829
    """

    sigma = NumberParam('sigma', low=0, low_open=True)

    def __init__(self, sigma=1., **lif_args):
        super(SoftLIFRate, self).__init__(**lif_args)
        self.sigma = sigma  # smoothing around the threshold

    @property
    def _argreprs(self):
        args = super(SoftLIFRate, self)._argreprs
        if self.sigma != 1.:
            args.append("sigma=%s" % self.sigma)
        return args

    def rates(self, x, gain, bias):
        J = gain * x + bias
        out = np.zeros_like(J)
        SoftLIFRate.step_math(self, dt=1, J=J, output=out)
        return out

    def step_math(self, dt, J, output):
        """Compute rates in Hz for input current (incl. bias)"""
        j = softplus(J - 1, sigma=self.sigma)
        output[:] = 0  # faster than output[j <= 0] = 0
        output[j > 0] = lif_j(j[j > 0], self.tau_ref, self.tau_rc,
                              amplitude=self.amplitude)

    def derivative(self, x, gain, bias):
        y = gain * x + bias - 1
        j = softplus(y, sigma=self.sigma)
        yy = y[j > 0]
        jj = j[j > 0]
        vv = lif_j(jj, self.tau_ref, self.tau_rc, amplitude=self.amplitude)

        d = np.zeros_like(j)
        d[j > 0] = (gain * self.tau_rc * vv * vv) / (
            self.amplitude * jj * (jj + 1) * (1 + np.exp(-yy / self.sigma)))
        return d


class FastLIF(nengo.neurons.LIF):
    """Faster version of the leaky integrate-and-fire (LIF) neuron model.

    This neuron model is faster than ``LIF`` but does not produce the ideal
    firing rate for larger ``dt`` due to linearization of the tuning curves.
    """

    def step_math(self, dt, J, spiked, voltage, refractory_time):

        # update voltage using accurate exponential integration scheme
        dV = -np.expm1(-dt / self.tau_rc) * (J - voltage)
        voltage += dV
        voltage[voltage < self.min_voltage] = self.min_voltage

        # update refractory period assuming no spikes for now
        refractory_time -= dt

        # set voltages of neurons still in their refractory period to 0
        # and linearly reduce voltage when partway out of ref. period
        voltage *= (1 - refractory_time / dt).clip(0, 1)

        # determine which neurons spike (if v > 1 set spiked = 1/dt, else 0)
        spiked[:] = (voltage > 1) / dt

        # linearly approximate time since neuron crossed spike threshold
        overshoot = (voltage[spiked > 0] - 1) / dV[spiked > 0]
        spiketime = dt * (1 - overshoot)

        # set spiking neurons' voltages to zero, and ref. time to tau_ref
        voltage[spiked > 0] = 0
        refractory_time[spiked > 0] = self.tau_ref + spiketime


def spikes2events(t, spikes):
    """Return an event-based representation of spikes (i.e. spike times)"""
    spikes = npext.array(spikes, copy=False, min_dims=2)
    if spikes.ndim > 2:
        raise ValidationError("Cannot handle %d-dimensional arrays"
                              % spikes.ndim, attr='spikes')
    if spikes.shape[-1] != len(t):
        raise ValidationError("Last dimension of 'spikes' must equal 'len(t)'",
                              attr='spikes')

    # find nonzero elements (spikes) in each row, and translate to times
    return [t[spike != 0] for spike in spikes]


def _rates_isi_events(t, events, midpoint, interp):
    import scipy.interpolate

    if len(events) == 0:
        return np.zeros_like(t)

    isis = np.diff(events)

    rt = np.zeros(len(events) + (1 if midpoint else 2))
    rt[1:-1] = 0.5*(events[:-1] + events[1:]) if midpoint else events
    rt[0], rt[-1] = t[0], t[-1]

    r = np.zeros_like(rt)
    r[1:len(isis) + 1] = 1. / isis

    f = scipy.interpolate.interp1d(rt, r, kind=interp, copy=False)
    return f(t)


def rates_isi(t, spikes, midpoint=False, interp='zero'):
    """Estimate firing rates from spikes using ISIs.

    Parameters
    ----------
    t : (M,) array_like
        The times at which raw spike data (spikes) is defined.
    spikes : (M, N) array_like
        The raw spike data from N neurons.
    midpoint : bool, optional
        If true, place interpolation points at midpoints of ISIs. Otherwise,
        the points are placed at the beginning of ISIs.
    interp : string, optional
        Interpolation type, passed to `scipy.interpolate.interp1d` as the
        ``kind`` parameter.

    Returns
    -------
    rates : (M, N) array_like
        The estimated neuron firing rates.
    """
    spike_times = spikes2events(t, spikes.T)
    rates = np.zeros(spikes.shape)
    for i, st in enumerate(spike_times):
        rates[:, i] = _rates_isi_events(t, st, midpoint, interp)

    return rates


def lowpass_filter(x, tau, kind='expon'):
    nt = x.shape[-1]

    if kind == 'expon':
        t = np.arange(0, 5 * tau)
        kern = np.exp(-t / tau) / tau
        delay = tau
    elif kind == 'gauss':
        std = tau / 2.
        t = np.arange(-4 * std, 4 * std)
        kern = np.exp(-0.5 * (t / std)**2) / np.sqrt(2 * np.pi * std**2)
        delay = 4 * std
    elif kind == 'alpha':
        alpha = 1. / tau
        t = np.arange(0, 5 * tau)
        kern = alpha**2 * t * np.exp(-alpha * t)
        delay = tau
    else:
        raise ValidationError("Unrecognized filter kind '%s'" % kind, 'kind')

    delay = int(np.round(delay))
    return np.array(
        [np.convolve(kern, xx, mode='full')[delay:nt + delay] for xx in x])


def rates_kernel(t, spikes, kind='gauss', tau=0.04):
    """Estimate firing rates from spikes using a kernel.

    Parameters
    ----------
    t : (M,) array_like
        The times at which raw spike data (spikes) is defined.
    spikes : (M, N) array_like
        The raw spike data from N neurons.
    kind : str {'expon', 'gauss', 'expogauss', 'alpha'}, optional
        The type of kernel to use. 'expon' is an exponential kernel, 'gauss' is
        a Gaussian (normal) kernel, 'expogauss' is an exponential followed by
        a Gaussian, and 'alpha' is an alpha function kernel.
    tau : float
        The time constant for the kernel. The optimal value will depend on the
        firing rate of the neurons, with a longer tau preferred for lower
        firing rates. The default value of 0.04 works well across a wide range
        of firing rates.
    """
    spikes = spikes.T
    spikes = npext.array(spikes, copy=False, min_dims=2)
    if spikes.ndim > 2:
        raise ValidationError("Cannot handle %d-dimensional arrays"
                              % spikes.ndim, attr='spikes')
    if spikes.shape[-1] != len(t):
        raise ValidationError("Last dimension of 'spikes' must equal 'len(t)'",
                              attr='spikes')

    n, nt = spikes.shape
    dt = t[1] - t[0]

    tau_i = tau / dt
    kind = kind.lower()
    if kind == 'expogauss':
        rates = lowpass_filter(spikes, tau_i, kind='expon')
        rates = lowpass_filter(rates, tau_i / 4, kind='gauss')
    else:
        rates = lowpass_filter(spikes, tau_i, kind=kind)

    return rates.T
